from __future__ import annotations
from pulser.noise_model import NoiseModel
import logging
import sys
import pathlib
from typing import TYPE_CHECKING
import torch

if TYPE_CHECKING:
    from emu_base.base_classes.callback import Callback


class BackendConfig:
    """The base backend configuration.

    Args:
        observables: a list of callbacks to compute observables
        with_modulation: if True, run the sequence with hardware modulation
        noise_model: The pulser.NoiseModel to use in the simulation.
        interaction_matrix: When specified, override the interaction terms in the Hamiltonian.
            This corresponds to the $U_{ij}$ terms in the documentation. Must be symmetric.
        interaction_cutoff: set interaction coefficients smaller than this to 0.
            This can improve the memory profile of the application for some backends.
        log_level: The output verbosity. Should be one of the constants from logging.
        log_file: a path to a file where to store the log, instead of printing to stdout

    Examples:
        >>> observables = [BitStrings(400, 100)] #compute 100 bitstrings at 400ns
        >>> noise_model = pulser.noise_model.NoiseModel()
        >>> interaction_matrix = [[1 for _ in range(nqubits)] for _ in range(nqubits)]
        >>> interaction_cutoff = 2.0 #this will turn off all the above interactions again
        >>> log_level = logging.warn
    """

    def __init__(
        self,
        *,
        observables: list[Callback] | None = None,
        with_modulation: bool = False,
        noise_model: NoiseModel | None = None,
        interaction_matrix: list[list[float]] | None = None,
        interaction_cutoff: float = 0.0,
        log_level: int = logging.INFO,
        log_file: pathlib.Path | None = None,
    ):
        if observables is None:
            observables = []
        self.callbacks = (
            observables  # we can add other types of callbacks, and just stack them
        )
        self.with_modulation = with_modulation
        self.noise_model = noise_model

        if interaction_matrix is not None and not (
            isinstance(interaction_matrix, list)
            and isinstance(interaction_matrix[0], list)
            and isinstance(interaction_matrix[0][0], float)
        ):
            raise ValueError(
                "Interaction matrix must be provided as a Python list of lists of floats"
            )

        if interaction_matrix is not None:
            int_mat = torch.tensor(interaction_matrix)
            tol = 1e-10
            if not (
                int_mat.numel() != 0
                and torch.all(torch.isreal(int_mat))
                and int_mat.dim() == 2
                and int_mat.shape[0] == int_mat.shape[1]
                and torch.allclose(int_mat, int_mat.T, atol=tol)
                and torch.norm(torch.diag(int_mat)) < tol
            ):
                raise ValueError("Interaction matrix is not symmetric and zero diag")

        self.interaction_matrix = interaction_matrix
        self.interaction_cutoff = interaction_cutoff
        self.logger = logging.getLogger("global_logger")
        self.log_file = log_file
        self.log_level = log_level

        self.init_logging()

        if noise_model is not None and (
            noise_model.runs != 1
            or noise_model.samples_per_run != 1
            or noise_model.runs is not None
            or noise_model.samples_per_run is not None
        ):
            self.logger.warning(
                "Warning: The runs and samples_per_run values of the NoiseModel are ignored!"
            )

    def init_logging(self) -> None:
        if self.log_file is None:
            logging.basicConfig(
                level=self.log_level, format="%(message)s", stream=sys.stdout, force=True
            )  # default to stream = sys.stderr
        else:
            logging.basicConfig(
                level=self.log_level,
                format="%(message)s",
                filename=str(self.log_file),
                filemode="w",
                force=True,
            )



from typing import Callable
import torch

DEFAULT_MAX_KRYLOV_DIM: int = 100


class KrylovExpResult:
    def __init__(
        self,
        result: torch.Tensor,
        converged: bool,
        happy_breakdown: bool,
        iteration_count: int,
    ):
        assert (not happy_breakdown) or converged

        self.converged = converged
        self.happy_breakdown = happy_breakdown
        self.iteration_count = iteration_count
        self.result = result


def krylov_exp_impl(
    op: Callable[[torch.Tensor], torch.Tensor],
    v: torch.Tensor,
    is_hermitian: bool,  # note: complex-proportional to its adjoint is enough
    exp_tolerance: float,
    norm_tolerance: float,
    max_krylov_dim: int = DEFAULT_MAX_KRYLOV_DIM,
) -> KrylovExpResult:
    """
    Computes exp(op).v using either the Lanczos or Arnoldi algorithm,
    based on the `is_hermitian` flag.
    All inputs must be on the same device.

    Convergence is checked using the exponential of the "extended T matrix", a criterion
    described in "Expokit: A Software Package for Computing Matrix Exponentials"
    (https://www.maths.uq.edu.au/expokit/paper.pdf).
    """

    initial_norm = v.norm()

    lanczos_vectors = [v / initial_norm]
    T = torch.zeros(max_krylov_dim + 2, max_krylov_dim + 2, dtype=v.dtype)

    for j in range(max_krylov_dim):
        w = op(lanczos_vectors[-1])

        n = w.norm()

        k_start = max(0, j - 1) if is_hermitian else 0
        for k in range(k_start, j + 1):
            overlap = torch.tensordot(lanczos_vectors[k].conj(), w, dims=w.dim())
            T[k, j] = overlap
            w = w - overlap * lanczos_vectors[k]

        n2 = w.norm()
        T[j + 1, j] = n2

        if n2 < norm_tolerance:
            # Happy breakdown
            expd = torch.linalg.matrix_exp(T[: j + 1, : j + 1])
            result = initial_norm * sum(
                a * b for a, b in zip(expd[:, 0], lanczos_vectors)
            )
            return KrylovExpResult(
                result=result, converged=True, happy_breakdown=True, iteration_count=j + 1
            )

        lanczos_vectors.append(w / n2)

        # Compute exponential of extended T matrix
        T[j + 2, j + 1] = 1
        expd = torch.linalg.matrix_exp(T[: j + 3, : j + 3])

        # Local truncation error estimation
        err1 = abs(expd[j + 1, 0])
        err2 = abs(expd[j + 2, 0] * n)

        err = err1 if err1 < err2 else (err1 * err2 / (err1 - err2))

        if err < exp_tolerance:
            # Converged
            result = initial_norm * sum(
                a * b for a, b in zip(expd[: len(lanczos_vectors), 0], lanczos_vectors)
            )
            return KrylovExpResult(
                result=result,
                converged=True,
                happy_breakdown=False,
                iteration_count=j + 1,
            )

    result = initial_norm * sum(
        a * b for a, b in zip(expd[: len(lanczos_vectors), 0], lanczos_vectors)
    )
    return KrylovExpResult(
        result=result,
        converged=False,
        happy_breakdown=False,
        iteration_count=max_krylov_dim,
    )


def krylov_exp(
    op: Callable[[torch.Tensor], torch.Tensor],
    v: torch.Tensor,
    exp_tolerance: float,
    norm_tolerance: float,
    is_hermitian: bool = True,  # note: complex-proportional to its adjoint is enough
    max_krylov_dim: int = DEFAULT_MAX_KRYLOV_DIM,
) -> torch.Tensor:
    krylov_result = krylov_exp_impl(
        op,
        v,
        is_hermitian=is_hermitian,
        exp_tolerance=exp_tolerance,
        norm_tolerance=norm_tolerance,
        max_krylov_dim=max_krylov_dim,
    )

    if not krylov_result.converged:
        raise RecursionError(
            "exponentiation algorithm did not converge to precision in allotted number of steps."
        )

    return krylov_result.result



from __future__ import annotations
from typing import Any, Iterable
from abc import ABC, abstractmethod
from collections import Counter


class State(ABC):
    """
    Base class enforcing an API for quantum states.
    Each backend will implement its own type of state, and the
    below methods.
    """

    @abstractmethod
    def inner(self, other: State) -> float | complex:
        """
        Compute the inner product between this state and other.
        Note that self is the left state in the inner product,
        so this function is linear in other, and anti-linear in self

        Args:
            other: the other state

        Returns:
            inner product
        """
        pass

    @abstractmethod
    def sample(
        self, num_shots: int, p_false_pos: float = 0.0, p_false_neg: float = 0.0
    ) -> Counter[str]:
        """
        Sample bitstrings from the state, taking into account error rates.

        Args:
            num_shots: how many bitstrings to sample
            p_false_pos: the rate at which a 0 is read as a 1
            p_false_neg: the rate at which a 1 is read as a 0

        Returns:
            the measured bitstrings, by count
        """
        pass

    @abstractmethod
    def __add__(self, other: State) -> State:
        """
        Computes the sum of two states.

        Args:
            other: the other state

        Returns:
            the summed state
        """
        pass

    @abstractmethod
    def __rmul__(self, scalar: complex) -> State:
        """
        Scale the state by a scale factor.

        Args:
            scalar: the scale factor

        Returns:
            the scaled state
        """
        pass

    @staticmethod
    @abstractmethod
    def from_state_string(
        *, basis: Iterable[str], nqubits: int, strings: dict[str, complex], **kwargs: Any
    ) -> State:
        """
        Construct a state from the pulser abstract representation
        <https://www.notion.so/pasqal/Abstract-State-and-Operator-Definition>
        For a list of existing bases, see
        <https://pulser.readthedocs.io/en/stable/conventions.html>

        Args:
            basis: A tuple containing the basis states.
            nqubits: the number of qubits.
            strings: A dictionary mapping state strings to complex or floats amplitudes

        Returns:
            the state in whatever format the backend provides.

        Examples:
            >>> afm_string_state = {"rrr": 1.0 / math.sqrt(2), "ggg": 1.0 / math.sqrt(2)}
            >>> afm_state = State.from_state_string(
            >>>     basis=("r", "g"), nqubits=3, strings=afm_string_state
            >>> )
        """
        pass



from abc import ABC, abstractmethod
from typing import Any, Optional, TYPE_CHECKING
from enum import Enum, auto

from emu_base.base_classes.config import BackendConfig
from emu_base.base_classes.operator import Operator
from emu_base.base_classes.state import State

if TYPE_CHECKING:
    from emu_base.base_classes.results import Results


class AggregationType(Enum):
    """
    Defines how to combine multiple values from different simulation results.
    """

    MEAN = auto()  # statistics.fmean or list/matrix-wise equivalent
    BAG_UNION = auto()  # Counter.__add__


class Callback(ABC):
    def __init__(self, evaluation_times: set[int]):
        """
        The callback base class that can be subclassed to add new kinds of results
        to the Results object returned by the Backend

        Args:
            evaluation_times: the times at which to add a result to Results
        """
        self.evaluation_times = evaluation_times

    def __call__(
        self, config: BackendConfig, t: int, state: State, H: Operator, result: "Results"
    ) -> None:
        """
        This function is called after each time step performed by the emulator.
        By default it calls apply to compute a result and put it in `result`
        if `t` in `self.evaluation_times`.
        It can be overloaded to define any custom behaviour for a `Callback`.

        Args:
            config: the config object passed to the run method
            t: the current time in ns
            state: the current state
            H: the Hamiltonian at this time
            result: the results object
        """
        if t in self.evaluation_times:
            value_to_store = self.apply(config, t, state, H)
            result.store(callback=self, time=t, value=value_to_store)

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the observable, can be used to index into the Results object.
        Some Callbacks might have multiple instances, such as a callback to compute
        a fidelity on some given state. In that case, this method could make sure
        each instance has a unique name.

        Returns:
            the name of the callback
        """
        pass

    @abstractmethod
    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        """
        This method must be implemented by subclasses. The result of this method
        gets put in the Results object.

        Args:
            config: the config object passed to the run method
            t: the current time in ns
            state: the current state
            H: the Hamiltonian at this time

        Returns:
            the result to put in Results
        """
        pass

    @property
    def default_aggregation_type(self) -> Optional[AggregationType]:
        """
        Defines how to combine by default multiple values from different simulation results.
        None means no default, therefore aggregator function is always user-provided.
        """
        return None



import statistics
from typing import Any, Callable
import collections
from emu_base.base_classes.callback import AggregationType


_NUMERIC_TYPES = {int, float, complex}


def mean_aggregator(
    values: list[Any],
) -> (
    complex
    | float
    | list[complex]
    | list[float]
    | list[list[complex]]
    | list[list[float]]
):  # FIXME: support tuples?
    if values == []:
        raise ValueError("Cannot average 0 samples")

    element_type = type(values[0])

    if element_type in _NUMERIC_TYPES:
        return statistics.fmean(values)

    if element_type != list:
        raise NotImplementedError("Cannot average this type of data")

    if values[0] == []:
        raise ValueError("Cannot average list of empty lists")

    sub_element_type = type(values[0][0])

    if sub_element_type in _NUMERIC_TYPES:
        dim = len(values[0])
        return [statistics.fmean(value[i] for value in values) for i in range(dim)]

    if sub_element_type != list:  # FIXME: ABC.Iterable? Collection? subclass?
        raise ValueError(f"Cannot average list of lists of {sub_element_type}")

    if values[0][0] == []:
        raise ValueError("Cannot average list of matrices with no columns")

    if (sub_sub_element_type := type(values[0][0][0])) not in _NUMERIC_TYPES:
        raise ValueError(f"Cannot average list of matrices of {sub_sub_element_type}")

    dim1 = len(values[0])
    dim2 = len(values[0][0])
    return [
        [statistics.fmean(value[i][j] for value in values) for j in range(dim2)]
        for i in range(dim1)
    ]


def bag_union_aggregator(values: list[collections.Counter]) -> collections.Counter:
    return sum(values, start=collections.Counter())


aggregation_types_definitions: dict[AggregationType, Callable] = {
    AggregationType.MEAN: mean_aggregator,
    AggregationType.BAG_UNION: bag_union_aggregator,
}



from dataclasses import dataclass, field
from typing import Any, Callable, Optional
from pathlib import Path
import json
import logging
import torch

from emu_base.base_classes.callback import Callback, AggregationType
from emu_base.base_classes.aggregators import aggregation_types_definitions


class ResultsEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, torch.Tensor):
            return obj.tolist()
        return super().default(obj)


@dataclass
class Results:
    """
    This class contains emulation results. Since the results written by
    an emulator are defined through callbacks, the contents of this class
    are not known a-priori.
    """

    statistics: Any = None  # Backend-specific data

    _results: dict[str, dict[int, Any]] = field(default_factory=dict)
    _default_aggregation_types: dict[str, Optional[AggregationType]] = field(
        default_factory=dict
    )

    @classmethod
    def aggregate(
        cls,
        results_to_aggregate: list["Results"],
        **aggregator_functions: Callable[[Any], Any],
    ) -> "Results":
        if len(results_to_aggregate) == 0:
            raise ValueError("no results to aggregate")

        if len(results_to_aggregate) == 1:
            return results_to_aggregate[0]

        stored_callbacks = set(results_to_aggregate[0].get_result_names())

        if not all(
            set(results.get_result_names()) == stored_callbacks
            for results in results_to_aggregate
        ):
            raise ValueError(
                "Monte-Carlo results seem to provide from incompatible simulations: "
                "they do not all contain the same observables"
            )

        aggregated: Results = cls()

        for stored_callback in stored_callbacks:
            aggregation_type = aggregator_functions.get(
                stored_callback,
                results_to_aggregate[0].get_aggregation_type(stored_callback),
            )

            if aggregation_type is None:
                logging.getLogger("global_logger").warning(
                    f"Skipping aggregation of `{stored_callback}`"
                )
                continue

            aggregation_function: Any = (
                aggregation_type
                if callable(aggregation_type)
                else aggregation_types_definitions[aggregation_type]
            )

            evaluation_times = results_to_aggregate[0].get_result_times(stored_callback)
            if not all(
                results.get_result_times(stored_callback) == evaluation_times
                for results in results_to_aggregate
            ):
                raise ValueError(
                    "Monte-Carlo results seem to provide from incompatible simulations: "
                    "the callbacks are not stored at the same times"
                )

            aggregated._results[stored_callback] = {
                t: aggregation_function(
                    [result[stored_callback, t] for result in results_to_aggregate]
                )
                for t in evaluation_times
            }

        return aggregated

    def store(self, *, callback: Callback, time: Any, value: Any) -> None:
        self._results.setdefault(callback.name, {})

        if time in self._results[callback.name]:
            raise ValueError(
                f"A value is already stored for observable '{callback.name}' at time {time}"
            )

        self._results[callback.name][time] = value
        self._default_aggregation_types[callback.name] = callback.default_aggregation_type

    def __getitem__(self, key: Any) -> Any:
        if isinstance(key, tuple):
            # results["energy", t]
            callback_name, time = key

            if callback_name not in self._results:
                raise ValueError(
                    f"No value for observable '{callback_name}' has been stored"
                )

            if time not in self._results[callback_name]:
                raise ValueError(
                    f"No value stored at time {time} for observable '{callback_name}'"
                )

            return self._results[callback_name][time]

        # results["energy"][t]
        assert isinstance(key, str)
        callback_name = key
        if callback_name not in self._results:
            raise ValueError(f"No value for observable '{callback_name}' has been stored")

        return self._results[key]

    def get_result_names(self) -> list[str]:
        """
        get a list of results present in this object

        Args:

        Returns:
            list of results by name

        """
        return list(self._results.keys())

    def get_result_times(self, name: str) -> list[int]:
        """
        get a list of times for which the given result has been stored

        Args:
            name: name of the result to get times of

        Returns:
            list of times in ns

        """
        return list(self._results[name].keys())

    def get_result(self, name: str, time: int) -> Any:
        """
        get the given result at the given time

        Args:
            name: name of the result to get
            time: time in ns at which to get the result

        Returns:
            the result

        """
        return self._results[name][time]

    def get_aggregation_type(self, name: str) -> Optional[AggregationType]:
        return self._default_aggregation_types[name]

    def dump(self, file_path: Path) -> None:
        with file_path.open("w") as file_handle:
            json.dump(
                {
                    "observables": self._results,
                    "statistics": self.statistics,
                },
                file_handle,
                cls=ResultsEncoder,
            )



from typing import Callable, Optional


class BrentsRootFinder:
    def __init__(
        self,
        *,
        start: float,
        end: float,
        f_start: float,
        f_end: float,
        epsilon: float = 1e-6,
    ):
        self.epsilon = epsilon

        assert start <= end
        self.a, self.b = start, end
        self.fa = f_start
        self.fb = f_end

        assert self.fa * self.fb < 0, "Function root needs to be between a and b"

        # b has to be the better guess
        if abs(self.fa) < abs(self.fb):
            self.a, self.b = self.b, self.a
            self.fa, self.fb = self.fb, self.fa

        self.c = self.a
        self.d = self.c
        self.fc = self.fa

        self.bisection = True
        self.current_guess = self.b
        self.next_abscissa: Optional[float] = None

    def get_next_abscissa(self) -> float:
        if abs(self.fc - self.fa) < self.epsilon or abs(self.fc - self.fb) < self.epsilon:
            # Secant method
            dx = self.fb * (self.b - self.a) / (self.fa - self.fb)
        else:
            # Inverse quadratic interpolation
            s = self.fb / self.fa
            r = self.fb / self.fc
            t = self.fa / self.fc
            q = (t - 1) * (s - 1) * (r - 1)
            p = s * (t * (r - t) * (self.c - self.b) + (r - 1) * (self.b - self.a))
            dx = p / q

        # Use bisection instead of interpolation
        # if the interpolation is not within bounds.
        delta = abs(2 * self.epsilon * self.b)
        adx = abs(dx)
        delta_bc = abs(self.b - self.c)
        delta_cd = abs(self.c - self.d)
        delta_ab = self.a - self.b
        if (
            (adx >= abs(3 * delta_ab / 4) or dx * delta_ab < 0)
            or (self.bisection and adx >= delta_bc / 2)
            or (not self.bisection and adx >= delta_cd / 2)
            or (self.bisection and delta_bc < delta)
            or (not self.bisection and delta_cd < delta)
        ):
            dx = (self.a - self.b) / 2
            self.bisection = True
        else:
            self.bisection = False

        self.next_abscissa = self.b + dx
        self.d = self.c
        self.c, self.fc = self.b, self.fb

        return self.next_abscissa

    def provide_ordinate(self, abscissa: float, ordinate: float) -> None:
        # First argument is just a safety
        assert (
            self.next_abscissa is not None and abscissa == self.next_abscissa
        ), "Something went wrong"

        # Update interval
        if self.fa * ordinate < 0:
            self.b, self.fb = abscissa, ordinate
        else:
            self.a, self.fa = abscissa, ordinate

        # b has to be the better guess
        if abs(self.fa) < abs(self.fb):
            self.a, self.b = self.b, self.a
            self.fa, self.fb = self.fb, self.fa

        self.current_guess = self.b

    def is_converged(self, tolerance: float) -> bool:
        return abs(self.b - self.a) < tolerance


def find_root_brents(
    f: Callable[[float], float],
    *,
    start: float,
    end: float,
    f_start: Optional[float] = None,
    f_end: Optional[float] = None,
    tolerance: float = 1e-6,
    epsilon: float = 1e-6,
) -> float:
    """
    Approximates and returns the zero of a scalar function using Brent's method.
    """
    f_start = f_start if f_start is not None else f(start)
    f_end = f_end if f_end is not None else f(end)

    root_finder = BrentsRootFinder(
        start=start, end=end, f_start=f_start, f_end=f_end, epsilon=epsilon
    )

    while not root_finder.is_converged(tolerance):
        x = root_finder.get_next_abscissa()
        root_finder.provide_ordinate(x, f(x))

    return root_finder.current_guess



import warnings
from abc import ABC, abstractmethod

from pulser import Sequence

from emu_base.base_classes.config import BackendConfig
from emu_base.base_classes.results import Results


class Backend(ABC):
    """
    Base class for different emulation backends.
    Forces backends to implement a run method.
    """

    @staticmethod
    def validate_sequence(sequence: Sequence) -> None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)

        if not isinstance(sequence, Sequence):
            raise TypeError(
                "The provided sequence has to be a valid " "pulser.Sequence instance."
            )
        if sequence.is_parametrized() or sequence.is_register_mappable():
            raise ValueError(
                "Not supported"
                "The provided sequence needs to be built to be simulated. Call"
                " `Sequence.build()` with the necessary parameters."
            )
        if not sequence._schedule:
            raise ValueError("The provided sequence has no declared channels.")
        if all(sequence._schedule[x][-1].tf == 0 for x in sequence.declared_channels):
            raise ValueError("No instructions given for the channels in the sequence.")

    @abstractmethod
    def run(self, sequence: Sequence, config: BackendConfig) -> Results:
        """
        Emulates the given sequence.

        Args:
            sequence: a Pulser sequence to simulate
            config: the config. Should be of the appropriate type for the backend

        Returns:
            the simulation results
        """
        pass



from .base_classes.results import Results
from .base_classes.callback import Callback, AggregationType
from .base_classes.config import BackendConfig
from .base_classes.operator import Operator
from .base_classes.state import State
from .base_classes.backend import Backend
from .base_classes.default_callbacks import (
    BitStrings,
    CorrelationMatrix,
    Energy,
    EnergyVariance,
    Expectation,
    Fidelity,
    QubitDensity,
    StateResult,
    SecondMomentOfEnergy,
)
from .constants import DEVICE_COUNT
from .pulser_adapter import PulserData, HamiltonianType
from .math.brents_root_finding import find_root_brents
from .math.krylov_exp import krylov_exp, DEFAULT_MAX_KRYLOV_DIM

__all__ = [
    "__version__",
    "Results",
    "BackendConfig",
    "Operator",
    "State",
    "Backend",
    "AggregationType",
    "Callback",
    "StateResult",
    "BitStrings",
    "QubitDensity",
    "CorrelationMatrix",
    "Expectation",
    "Fidelity",
    "Energy",
    "EnergyVariance",
    "SecondMomentOfEnergy",
    "PulserData",
    "find_root_brents",
    "krylov_exp",
    "HamiltonianType",
    "DEFAULT_MAX_KRYLOV_DIM",
    "DEVICE_COUNT",
]

__version__ = "1.2.7"



from pulser.noise_model import NoiseModel
import torch
import math


def get_lindblad_operators(
    *, noise_type: str, noise_model: NoiseModel
) -> list[torch.Tensor]:
    assert noise_type in noise_model.noise_types

    if noise_type == "relaxation":
        c = math.sqrt(noise_model.relaxation_rate)
        return [
            torch.tensor([[0, c], [0, 0]], dtype=torch.complex128),
        ]

    if noise_type == "dephasing":
        if noise_model.hyperfine_dephasing_rate != 0.0:
            raise NotImplementedError("hyperfine_dephasing_rate is unsupported")

        c = math.sqrt(noise_model.dephasing_rate / 2)
        return [
            torch.tensor([[-c, 0], [0, c]], dtype=torch.complex128),
        ]

    if noise_type == "depolarizing":
        c = math.sqrt(noise_model.depolarizing_rate / 4)
        return [
            torch.tensor([[0, c], [c, 0]], dtype=torch.complex128),
            torch.tensor([[0, 1j * c], [-1j * c, 0]], dtype=torch.complex128),
            torch.tensor([[-c, 0], [0, c]], dtype=torch.complex128),
        ]

    if noise_type == "eff_noise":
        if not all(
            isinstance(op, torch.Tensor) and op.shape == (2, 2)
            for op in noise_model.eff_noise_opers
        ):
            raise ValueError("Only 2 * 2 effective noise operator matrices are supported")

        return [
            math.sqrt(rate)
            * torch.flip(op if isinstance(op, torch.Tensor) else torch.tensor(op), (0, 1))
            for rate, op in zip(noise_model.eff_noise_rates, noise_model.eff_noise_opers)
        ]

    raise ValueError(f"Unknown noise type: {noise_type}")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable

from emu_base.base_classes.state import State


QuditOp = dict[str, complex]  # single qubit operator
TensorOp = list[tuple[QuditOp, list[int]]]  # QuditOp applied to list of qubits
FullOp = list[tuple[complex, TensorOp]]  # weighted sum of TensorOp


class Operator(ABC):
    @abstractmethod
    def __mul__(self, other: State) -> State:
        """
        Apply the operator to a state

        Args:
            other: the state to apply this operator to

        Returns:
            the resulting state
        """
        pass

    @abstractmethod
    def __add__(self, other: Operator) -> Operator:
        """
        Computes the sum of two operators.

        Args:
            other: the other operator

        Returns:
           the summed operator
        """
        pass

    @abstractmethod
    def expect(self, state: State) -> float | complex:
        """
        Compute the expectation value of self on the given state.

        Args:
            state: the state with which to compute

        Returns:
            the expectation
        """

    @staticmethod
    @abstractmethod
    def from_operator_string(
        basis: Iterable[str],
        nqubits: int,
        operations: FullOp,
        operators: dict[str, QuditOp] = {},
        /,
        **kwargs: Any,
    ) -> Operator:
        """
        Create an operator in the backend-specific format from the
        pulser abstract representation
        <https://www.notion.so/pasqal/Abstract-State-and-Operator-Definition>
        By default it supports strings 'ij', where i and j in basis,
        to denote |i><j|, but additional symbols can be defined in operators
        For a list of existing bases, see
        <https://pulser.readthedocs.io/en/stable/conventions.html>

        Args:
            basis: the eigenstates in the basis to use
            nqubits: how many qubits there are in the state
            operations: which bitstrings make up the state with what weight
            operators: additional symbols to be used in operations

        Returns:
            the operator in whatever format the backend provides.

        Examples:
            >>> basis = {"r", "g"} #rydberg basis
            >>> nqubits = 3 #or whatever
            >>> x = {"rg": 1.0, "gr": 1.0}
            >>> z = {"gg": 1.0, "rr": -1.0}
            >>> operators = {"X": x, "Z": z} #define X and Z as conveniences
            >>>
            >>> operations = [ # 4 X1X + 3 1Z1
            >>>     (
            >>>         1.0,
            >>>         [
            >>>             ({"X": 2.0}, [0, 2]),
            >>>             ({"Z": 3.0}, [1]),
            >>>         ],
            >>>     )
            >>> ]
            >>> op = Operator.from_operator_string(basis, nqubits, operations, operators)
        """
        pass

    @abstractmethod
    def __rmul__(self, scalar: complex) -> Operator:
        """
        Scale the operator by a scale factor.

        Args:
            scalar: the scale factor

        Returns:
            the scaled operator
        """
        pass

    @abstractmethod
    def __matmul__(self, other: Operator) -> Operator:
        """
        Compose two operators. The ordering is that
        self is applied after other.

        Args:
            other: the operator to compose with self

        Returns:
            the composed operator
        """
        pass



from .operator import Operator
from .state import State
from .results import Results
from .callback import Callback
from .default_callbacks import (
    StateResult,
    BitStrings,
    QubitDensity,
    CorrelationMatrix,
    Expectation,
    Fidelity,
    Energy,
    EnergyVariance,
    SecondMomentOfEnergy,
)

__all__ = [
    "Operator",
    "State",
    "Results",
    "Callback",
    "StateResult",
    "BitStrings",
    "QubitDensity",
    "CorrelationMatrix",
    "Expectation",
    "Fidelity",
    "Energy",
    "EnergyVariance",
    "SecondMomentOfEnergy",
]



from copy import deepcopy
from typing import Any

from emu_base.base_classes.callback import Callback, AggregationType
from emu_base.base_classes.config import BackendConfig
from emu_base.base_classes.operator import Operator
from emu_base.base_classes.state import State


class StateResult(Callback):
    """
    Store the quantum state in whatever format the backend provides

    Args:
        evaluation_times: the times at which to store the state
    """

    def __init__(self, evaluation_times: set[int]):
        super().__init__(evaluation_times)

    name = "state"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        return deepcopy(state)


class BitStrings(Callback):
    """
    Store bitstrings sampled from the current state. Error rates are taken from the config
    passed to the run method of the backend. The bitstrings are stored as a Counter[str].

    Args:
        evaluation_times: the times at which to sample bitstrings
        num_shots: how many bitstrings to sample each time this observable is computed
    """

    def __init__(self, evaluation_times: set[int], num_shots: int = 1000):
        super().__init__(evaluation_times)
        self.num_shots = num_shots

    name = "bitstrings"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        p_false_pos = (
            0.0 if config.noise_model is None else config.noise_model.p_false_pos
        )
        p_false_neg = (
            0.0 if config.noise_model is None else config.noise_model.p_false_neg
        )

        return state.sample(self.num_shots, p_false_pos, p_false_neg)

    default_aggregation_type = AggregationType.BAG_UNION


_fidelity_counter = -1


class Fidelity(Callback):
    """
    Store $<ψ|φ(t)>$ for the given state $|ψ>$,
    and the state $|φ(t)>$ obtained by time evolution.

    Args:
        evaluation_times: the times at which to compute the fidelity
        state: the state |ψ>. Note that this must be of appropriate type for the backend

    Examples:
        >>> state = State.from_state_string(...) #see State API
        >>> fidelity = Fidelity([400], state) #measure fidelity on state at t=400ns
    """

    def __init__(self, evaluation_times: set[int], state: State):
        super().__init__(evaluation_times)
        global _fidelity_counter
        _fidelity_counter += 1
        self.index = _fidelity_counter
        self.state = state

    @property
    def name(self) -> str:
        return f"fidelity_{self.index}"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        return self.state.inner(state)


_expectation_counter = -1


class Expectation(Callback):
    """
    Store the expectation of the given operator on the current state
    (i.e. $\\langle φ(t)|\\mathrm{operator}|φ(t)\\rangle$).

    Args:
        evaluation_times: the times at which to compute the expectation
        operator: the operator to measure. Must be of appropriate type for the backend.

    Examples:
        >>> op = Operator.from_operator_string(...) #see Operator API
        >>> expectation = Expectation([400], op) #measure the expecation of op at t=400ns
    """

    def __init__(self, evaluation_times: set[int], operator: Operator):
        super().__init__(evaluation_times)
        global _expectation_counter
        _expectation_counter += 1
        self.index = _expectation_counter
        self.operator = operator

    @property
    def name(self) -> str:
        return f"expectation_{self.index}"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        return self.operator.expect(state)

    default_aggregation_type = AggregationType.MEAN


class CorrelationMatrix(Callback):
    """
    Store the correlation matrix for the current state.
    Requires specification of the basis used in the emulation
    https://pulser.readthedocs.io/en/stable/conventions.html
    It currently supports
    - the rydberg basis ('r','g')
    - the xy basis ('0', '1')
    and returns

    `[[<φ(t)|n_i n_j|φ(t)> for j in qubits] for i in qubits]`

    n_i being the operator that projects qubit i onto the state that measures as 1.
    The diagonal of this matrix is the QubitDensity. The correlation matrix
    is stored as a list of lists.

    Args:
        evaluation_times: the times at which to compute the correlation matrix
        basis: the basis used by the sequence
        nqubits: the number of qubits in the Register

    Notes:
        See the API for `Operator.from_operator_string` for an example of what to do with
        basis and nqubits.
    """

    def __init__(self, evaluation_times: set[int], basis: tuple[str, ...], nqubits: int):
        super().__init__(evaluation_times)
        self.operators: list[list[Operator]] | None = None
        self.basis = set(basis)
        if self.basis == {"r", "g"}:
            self.op_string = "rr"
        elif self.basis == {"0", "1"}:
            self.op_string = "11"
        else:
            raise ValueError("Unsupported basis provided")
        self.nqubits = nqubits

    name = "correlation_matrix"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        if hasattr(state, "get_correlation_matrix") and callable(
            state.get_correlation_matrix
        ):
            return state.get_correlation_matrix()

        if self.operators is None or not isinstance(self.operators[0], type(H)):
            self.operators = [
                [
                    H.from_operator_string(
                        self.basis,
                        self.nqubits,
                        [(1.0, [({self.op_string: 1.0}, list({i, j}))])],
                    )
                    for j in range(self.nqubits)
                ]
                for i in range(self.nqubits)
            ]
        return [[op.expect(state).real for op in ops] for ops in self.operators]

    default_aggregation_type = AggregationType.MEAN


class QubitDensity(Callback):
    """
    Requires specification of the basis used in the emulation
    https://pulser.readthedocs.io/en/stable/conventions.html
    It currently supports
    - the rydberg basis ('r','g')
    - the xy basis ('0', '1')
    and returns

    `[<φ(t)|n_i|φ(t)> for i in qubits]`

    n_i being the operator that projects qubit i onto the state that measures as 1.
    The qubit density is stored as a list.

    Args:
        evaluation_times: the times at which to compute the density
        basis: the basis used by the sequence
        nqubits: the number of qubits in the Register

    Notes:
        See the API for `State.from_state_string` for an example of what to do with
        basis and nqubits.
    """

    def __init__(self, evaluation_times: set[int], basis: tuple[str, ...], nqubits: int):
        super().__init__(evaluation_times)
        self.operators: list[Operator] | None = None
        self.basis = set(basis)
        if self.basis == {"r", "g"}:
            self.op_string = "rr"
        elif self.basis == {"0", "1"}:
            self.op_string = "11"
        else:
            raise ValueError("Unsupported basis provided")
        self.nqubits = nqubits

    name = "qubit_density"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        if self.operators is None or not isinstance(self.operators[0], type(H)):
            self.operators = [
                H.from_operator_string(
                    self.basis, self.nqubits, [(1.0, [({self.op_string: 1.0}, [i])])]
                )
                for i in range(self.nqubits)
            ]
        return [op.expect(state).real for op in self.operators]

    default_aggregation_type = AggregationType.MEAN


class Energy(Callback):
    """
    Store the expectation value of the current Hamiltonian
    (i.e. $\\langle φ(t)|H(t)|φ(t) \\rangle$)

    Args:
        evaluation_times: the times at which to compute the expectation
    """

    def __init__(self, evaluation_times: set[int]):
        super().__init__(evaluation_times)

    name = "energy"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        return H.expect(state).real

    default_aggregation_type = AggregationType.MEAN


class EnergyVariance(Callback):
    """
    Store the variance of the current Hamiltonian
    (i.e. $\\langle φ(t)|H(t)^2|φ(t)\\rangle - \\langle φ(t)|H(t)|φ(t)\\rangle^2$)

    Args:
        evaluation_times: the times at which to compute the variance
    """

    def __init__(self, evaluation_times: set[int]):
        super().__init__(evaluation_times)

    name = "energy_variance"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        h_squared = H @ H
        return h_squared.expect(state).real - H.expect(state).real ** 2

    # Explicitely setting this to None out of safety: in the case of MonteCarlo,
    # the aggregated variance cannot be computed from this callback.
    # Instead, one first need to average Energy and SecondMomentOfEnergy,
    # and then compute the variance with the formula:
    # AggregatedEnergyVariance = AveragedSecondMomentOfEnergy - AveragedEnergy**2
    default_aggregation_type = None


class SecondMomentOfEnergy(Callback):
    """
    Store the expectation value $\\langle φ(t)|H(t)^2|φ(t)\\rangle$.
    Useful for computing the variance when averaging over many executions of the program.

    Args:
        evaluation_times: the times at which to compute the variance
    """

    def __init__(self, evaluation_times: set[int]):
        super().__init__(evaluation_times)

    name = "second_moment_of_energy"

    def apply(self, config: BackendConfig, t: int, state: State, H: Operator) -> Any:
        h_squared = H @ H
        return h_squared.expect(state).real

    default_aggregation_type = AggregationType.MEAN



import pulser
from typing import Tuple
import torch
import math
from pulser.noise_model import NoiseModel
from enum import Enum

from pulser.register.base_register import BaseRegister

from emu_base.base_classes.config import BackendConfig
from emu_base.lindblad_operators import get_lindblad_operators
from emu_base.utils import dist2, dist3


class HamiltonianType(Enum):
    Rydberg = 1
    XY = 2


def _get_qubit_positions(
    register: BaseRegister,
) -> list[torch.Tensor]:
    """Conversion from pulser Register to emu-mps register (torch type).
    Each element will be given as [Rx,Ry,Rz]"""

    positions = [position.as_tensor() for position in register.qubits.values()]

    if len(positions[0]) == 2:
        return [torch.cat((position, torch.zeros(1))) for position in positions]
    return positions


def _rydberg_interaction(sequence: pulser.Sequence) -> torch.Tensor:
    """
    Computes the Ising interaction matrix from the qubit positions.
    Hᵢⱼ=C₆/R⁶ᵢⱼ (nᵢ⊗ nⱼ)
    """

    num_qubits = len(sequence.register.qubit_ids)

    c6 = sequence.device.interaction_coeff

    qubit_positions = _get_qubit_positions(sequence.register)
    interaction_matrix = torch.zeros(num_qubits, num_qubits)

    for numi in range(len(qubit_positions)):
        for numj in range(numi + 1, len(qubit_positions)):
            interaction_matrix[numi][numj] = (
                c6 / dist2(qubit_positions[numi], qubit_positions[numj]) ** 3
            )
            interaction_matrix[numj, numi] = interaction_matrix[numi, numj]
    return interaction_matrix


def _xy_interaction(sequence: pulser.Sequence) -> torch.Tensor:
    """
    Computes the XY interaction matrix from the qubit positions.
    C₃ (1−3 cos(𝜃ᵢⱼ)²)/ Rᵢⱼ³ (𝜎ᵢ⁺ 𝜎ⱼ⁻ +  𝜎ᵢ⁻ 𝜎ⱼ⁺)
    """
    num_qubits = len(sequence.register.qubit_ids)

    c3 = sequence.device.interaction_coeff_xy

    qubit_positions = _get_qubit_positions(sequence.register)
    interaction_matrix = torch.zeros(num_qubits, num_qubits)
    mag_field = torch.tensor(sequence.magnetic_field)  # by default [0.0,0.0,30.0]
    mag_norm = torch.linalg.norm(mag_field)

    for numi in range(len(qubit_positions)):
        for numj in range(numi + 1, len(qubit_positions)):
            cosine = 0
            if mag_norm >= 1e-8:  # selected by hand
                cosine = torch.dot(
                    (qubit_positions[numi] - qubit_positions[numj]), mag_field
                ) / (
                    torch.linalg.norm(qubit_positions[numi] - qubit_positions[numj])
                    * mag_norm
                )

            interaction_matrix[numi][numj] = (
                c3  # check this value with pulser people
                * (1 - 3 * cosine**2)
                / dist3(qubit_positions[numi], qubit_positions[numj])
            )
            interaction_matrix[numj, numi] = interaction_matrix[numi, numj]

    return interaction_matrix


def _extract_omega_delta_phi(
    *,
    sequence: pulser.Sequence,
    target_times: list[int],
    with_modulation: bool,
    laser_waist: float | None,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Samples the Pulser sequence and returns a tuple of tensors (omega, delta, phi)
    containing:
    - omega[i, q] = amplitude at time i * dt for qubit q
    - delta[i, q] = detuning at time i * dt for qubit q
    - phi[i, q] = phase at time i * dt for qubit q

    if laser_waist is w_0 != None, the omega values coming from the global pulse channel
    will me modulated as $\\Omega_i=\\Omega_i e^{-r_i^2/w_0^2}$
    """

    if with_modulation and sequence._slm_mask_targets:
        raise NotImplementedError(
            "Simulation of sequences combining an SLM mask and output "
            "modulation is not supported."
        )

    samples = pulser.sampler.sample(
        sequence,
        modulation=with_modulation,
        extended_duration=sequence.get_duration(include_fall_time=with_modulation),
    )
    sequence_dict = samples.to_nested_dict(all_local=True, samples_type="tensor")["Local"]

    if "ground-rydberg" in sequence_dict and len(sequence_dict) == 1:
        locals_a_d_p = sequence_dict["ground-rydberg"]
    elif "XY" in sequence_dict and len(sequence_dict) == 1:
        locals_a_d_p = sequence_dict["XY"]
    else:
        raise ValueError("Emu-MPS only accepts ground-rydberg or mw_global channels")

    max_duration = sequence.get_duration(include_fall_time=with_modulation)

    nsamples = len(target_times) - 1
    omega = torch.zeros(
        nsamples,
        len(sequence.register.qubit_ids),
        dtype=torch.complex128,
    )

    delta = torch.zeros(
        nsamples,
        len(sequence.register.qubit_ids),
        dtype=torch.complex128,
    )
    phi = torch.zeros(
        nsamples,
        len(sequence.register.qubit_ids),
        dtype=torch.complex128,
    )

    if laser_waist:
        qubit_positions = _get_qubit_positions(sequence.register)
        waist_factors = torch.tensor(
            [math.exp(-((x[:2].norm() / laser_waist) ** 2)) for x in qubit_positions]
        )
    else:
        waist_factors = torch.ones(len(sequence.register.qubit_ids))

    global_times = set()
    for ch, ch_samples in samples.channel_samples.items():
        if samples._ch_objs[ch].addressing == "Global":
            for slot in ch_samples.slots:
                global_times |= set(i for i in range(slot.ti, slot.tf))

    omega_1 = torch.zeros_like(omega[0])
    omega_2 = torch.zeros_like(omega[0])

    for i in range(nsamples):
        t = (target_times[i] + target_times[i + 1]) / 2
        # The sampled values correspond to the start of each interval
        # To maximize the order of the solver, we need the values in the middle
        if math.ceil(t) < max_duration:
            # If we're not the final step, approximate this using linear interpolation
            # Note that for dt even, t1=t2
            for q_pos, q_id in enumerate(sequence.register.qubit_ids):
                t1 = math.floor(t)
                t2 = math.ceil(t)
                omega_1[q_pos] = locals_a_d_p[q_id]["amp"][t1]
                omega_2[q_pos] = locals_a_d_p[q_id]["amp"][t2]
                delta[i, q_pos] = (
                    locals_a_d_p[q_id]["det"][t1] + locals_a_d_p[q_id]["det"][t2]
                ) / 2.0
                phi[i, q_pos] = (
                    locals_a_d_p[q_id]["phase"][t1] + locals_a_d_p[q_id]["phase"][t2]
                ) / 2.0
            # omegas at different times need to have the laser waist applied independently
            if t1 in global_times:
                omega_1 *= waist_factors
            if t2 in global_times:
                omega_2 *= waist_factors
            omega[i] = 0.5 * (omega_1 + omega_2)
        else:
            # We're in the final step and dt=1, approximate this using linear extrapolation
            # we can reuse omega_1 and omega_2 from before
            for q_pos, q_id in enumerate(sequence.register.qubit_ids):
                delta[i, q_pos] = (
                    3.0 * locals_a_d_p[q_id]["det"][t2] - locals_a_d_p[q_id]["det"][t1]
                ) / 2.0
                phi[i, q_pos] = (
                    3.0 * locals_a_d_p[q_id]["phase"][t2]
                    - locals_a_d_p[q_id]["phase"][t1]
                ) / 2.0
            omega[i] = torch.clamp(0.5 * (3 * omega_2 - omega_1).real, min=0.0)

    return omega, delta, phi


_NON_LINDBLADIAN_NOISE = {"SPAM", "doppler", "amplitude"}


def _get_all_lindblad_noise_operators(
    noise_model: NoiseModel | None,
) -> list[torch.Tensor]:
    if noise_model is None:
        return []

    return [
        op
        for noise_type in noise_model.noise_types
        if noise_type not in _NON_LINDBLADIAN_NOISE
        for op in get_lindblad_operators(noise_type=noise_type, noise_model=noise_model)
    ]


class PulserData:
    slm_end_time: float
    full_interaction_matrix: torch.Tensor
    masked_interaction_matrix: torch.Tensor
    omega: torch.Tensor
    delta: torch.Tensor
    phi: torch.Tensor
    hamiltonian_type: HamiltonianType
    lindblad_ops: list[torch.Tensor]

    def __init__(self, *, sequence: pulser.Sequence, config: BackendConfig, dt: int):
        self.qubit_count = len(sequence.register.qubit_ids)

        # the end value is exclusive, so add +1
        observable_times = set(torch.arange(0, sequence.get_duration() + 1, dt).tolist())
        observable_times.add(sequence.get_duration())
        for obs in config.callbacks:
            observable_times |= set(obs.evaluation_times)
        self.target_times = list(observable_times)
        self.target_times.sort()

        laser_waist = (
            config.noise_model.laser_waist if config.noise_model is not None else None
        )
        self.omega, self.delta, self.phi = _extract_omega_delta_phi(
            sequence=sequence,
            target_times=self.target_times,
            with_modulation=config.with_modulation,
            laser_waist=laser_waist,
        )
        self.lindblad_ops = _get_all_lindblad_noise_operators(config.noise_model)
        self.has_lindblad_noise: bool = self.lindblad_ops != []

        addressed_basis = sequence.get_addressed_bases()[0]
        if addressed_basis == "ground-rydberg":  # for local and global
            self.hamiltonian_type = HamiltonianType.Rydberg
        elif addressed_basis == "XY":
            self.hamiltonian_type = HamiltonianType.XY
        else:
            raise ValueError(f"Unsupported basis: {addressed_basis}")

        if config.interaction_matrix is not None:
            assert len(config.interaction_matrix) == self.qubit_count, (
                "The number of qubits in the register should be the same as the size of "
                "the interaction matrix"
            )

            self.full_interaction_matrix = torch.tensor(
                config.interaction_matrix, dtype=torch.float64
            )
        elif self.hamiltonian_type == HamiltonianType.Rydberg:
            self.full_interaction_matrix = _rydberg_interaction(sequence)
        elif self.hamiltonian_type == HamiltonianType.XY:
            self.full_interaction_matrix = _xy_interaction(sequence)
        self.full_interaction_matrix[
            torch.abs(self.full_interaction_matrix) < config.interaction_cutoff
        ] = 0.0
        self.masked_interaction_matrix = self.full_interaction_matrix.clone()

        self.slm_end_time = (
            sequence._slm_mask_time[1] if len(sequence._slm_mask_time) > 1 else 0.0
        )

        # disable interaction for SLM masked qubits
        slm_targets = list(sequence._slm_mask_targets)
        for target in sequence.register.find_indices(slm_targets):
            self.masked_interaction_matrix[target] = 0.0
            self.masked_interaction_matrix[:, target] = 0.0



import torch


def dist2(left: torch.Tensor, right: torch.Tensor) -> float:
    return torch.dist(left, right).item() ** 2


def dist3(left: torch.Tensor, right: torch.Tensor) -> float:
    return torch.dist(left, right).item() ** 3
